"""
init file for the 'hoppr.core_plugins.report_generator' module
"""

from hoppr.core_plugins.report_generator.report_generator import Report, ReportGenerator

__all__ = ["Report", "ReportGenerator"]



"""
Generate JSON schema files from pydantic models
"""
from __future__ import annotations

from pathlib import Path
from typing import Any

from ruamel.yaml import YAML

from hoppr.models import CredentialsFile, HopprBaseModel, HopprSchemaModel, ManifestFile, TransferFile


_yaml_writer = YAML(typ="safe", pure=True)
_yaml_writer.default_flow_style = False
_yaml_writer.map_indent = 2
_yaml_writer.sequence_indent = 4
_yaml_writer.sequence_dash_offset = 2
_yaml_writer.sort_base_mapping_type_on_output = False  # type: ignore[assignment]


def write_json_schema(schema_name: str, model_type: type[HopprBaseModel]) -> None:  # pragma: no cover
    """Write JSON schema to file."""
    schema_file = Path(f"hoppr-{schema_name}-schema-v1.json")
    schema_file.write_text(data=model_type.schema_json(indent=2), encoding="utf-8")


def write_yaml_schema(schema_name: str, model_type: type[HopprBaseModel]) -> None:  # pragma: no cover
    """Write YAML schema to file."""
    schema_file = Path(f"hoppr-{schema_name}-schema-v1.yml")

    _yaml_writer.dump(data=model_type.schema(by_alias=True), stream=schema_file)

    # Replace generated single quotes with double
    content = schema_file.read_text(encoding="utf-8")
    content = content.replace("'", '"')
    content = f"---\n{content}"
    schema_file.write_text(content, encoding="utf-8")


if __name__ == "__main__":
    schema_args: list[dict[str, Any]] = [
        {"schema_name": "credentials", "model_type": CredentialsFile},
        {"schema_name": "manifest", "model_type": ManifestFile},
        {"schema_name": "transfer", "model_type": TransferFile},
        {"schema_name": "combined", "model_type": HopprSchemaModel},
    ]

    for args in schema_args:
        # Write JSON schema files
        write_json_schema(**args)

        # Write YAML schema files
        write_yaml_schema(**args)



"""
Collector plugin for Gem packages
"""
from __future__ import annotations

import jmespath
import requests

from requests import HTTPError

import hoppr.net
import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectGemPlugin(SerialCollectorPlugin):
    """
    Collector plugin for Gem packages
    """

    supported_purl_types = ["gem"]
    products: list[str] = ["gem/*"]
    system_repositories: list[str] = ["https://rubygems.org/api/v2/rubygems"]

    def get_version(self):
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)

        api_url = RepositoryUrl(url=repo_url) / purl.name / "versions" / f"{purl.version}.json"

        if platform := purl.qualifiers.get("platform"):
            api_url = RepositoryUrl(url=f"{api_url}?platform={platform}")

        response = requests.get(str(api_url), timeout=60)

        try:
            response.raise_for_status()
        except HTTPError:
            msg = f"RubyGems API web request error, request URL='{api_url}', status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        download_url: str | None = jmespath.search(expression="gem_uri", data=response.json())

        if not download_url:
            msg = f"Unable to retrieve download URL for Gem '{comp.purl}'"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        self.get_logger().info(msg=f"Copying Gem package from {api_url}", indent_level=2)

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        self.get_logger().info(msg=target_dir.as_posix())

        package_out_path = target_dir / f"{purl.name}-{purl.version}.gem"

        self.get_logger().info(msg="Downloading Gem:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {package_out_path}", indent_level=3)

        response = hoppr.net.download_file(url=download_url, dest=str(package_out_path))

        result = Result.from_http_response(response=response, return_obj=comp)

        if not result.is_success():
            msg = f"Failed to download Gem package {comp.purl}, status_code={response.status_code}"
            self.get_logger().error(msg=msg, indent_level=2)
            return Result.fail(message=msg)

        self.set_collection_params(comp, repo_url, target_dir)
        return Result.success(return_obj=comp)



"""
Base model for Hoppr config files
"""
from __future__ import annotations

import logging

from pathlib import Path
from threading import _RLock as RLock

from pydantic import Field

from hoppr.models.base import HopprBaseModel
from hoppr.models.credentials import CredentialRequiredService, Credentials, CredentialsFile
from hoppr.models.manifest import Manifest, ManifestFile, Repositories
from hoppr.models.sbom import Sbom
from hoppr.models.transfer import StageRef, Transfer, TransferFile

__all__ = [
    "Credentials",
    "CredentialsFile",
    "HopprBaseModel",
    "Manifest",
    "ManifestFile",
    "Transfer",
    "TransferFile",
]


class HopprSchemaModel(HopprBaseModel):
    """
    Consolidated Hoppr config file schema definition
    """

    __root__: CredentialsFile | ManifestFile | TransferFile = Field(..., discriminator="kind")


class HopprContext(HopprBaseModel):
    """
    Consolidated data model for Hoppr runtime context
    """

    collect_root_dir: Path
    consolidated_sbom: Sbom
    credential_required_services: list[CredentialRequiredService] | None
    delivered_sbom: Sbom
    log_level: int = logging.INFO
    logfile_location: Path = Path("hoppr.log")
    logfile_lock: RLock
    max_attempts: int = 3
    max_processes: int
    previous_delivery: Path | str | None = None
    repositories: Repositories
    retry_wait_seconds: float = 5
    sboms: list[Sbom]
    stages: list[StageRef]
    strict_repos: bool = True


HopprContext.update_forward_refs()



"""
Hoppr stage output Rich layout
"""
from __future__ import annotations

import warnings

from typing import TYPE_CHECKING, Final, MutableMapping

from rich.console import Console
from rich.containers import Renderables
from rich.layout import Layout
from rich.panel import Panel
from rich.progress import BarColumn, Progress, SpinnerColumn, Task, TaskID, TextColumn
from rich.segment import SegmentLines
from rich.table import Column, Table
from rich.text import Text
from typing_extensions import override

import hoppr.utils

if TYPE_CHECKING:
    from rich.console import ConsoleOptions, RenderableType, RenderResult
    from rich.style import StyleType
    from rich.text import TextType


# Renderable constants
BORDER_LINES = 2
HEADER_SIZE = 3

# Padding constants
PAD_BOTTOM = 0
PAD_LEFT = 1
PAD_RIGHT = 1
PAD_TOP = 0

# Panel size ratio constants
JOBS_PANEL_RATIO_SIZE = 1
OUTPUT_PANEL_RATIO_SIZE = 2
TOTAL_PANEL_RATIO_SIZE = JOBS_PANEL_RATIO_SIZE + OUTPUT_PANEL_RATIO_SIZE
OUTPUT_PANEL_RATIO = OUTPUT_PANEL_RATIO_SIZE / TOTAL_PANEL_RATIO_SIZE

console = Console()


class HopprBasePanel(Panel):  # pylint: disable=too-few-public-methods
    """
    Customized base Rich Panel
    """

    def __init__(self, renderable: RenderableType, title: str | None = None, **kwargs) -> None:
        super().__init__(
            renderable=renderable,
            padding=(PAD_TOP, PAD_RIGHT, PAD_BOTTOM, PAD_LEFT),
            title=f"[bold blue]{title}" if title else None,
            title_align="left",
            border_style="bold purple",
            style="white on grey0",
            **kwargs,
        )


class HopprConsole(Renderables):  # pylint: disable=too-few-public-methods
    """
    Renderable Rich Console
    """

    def __rich_console__(  # pylint: disable=arguments-renamed
        self, console_: Console, options: ConsoleOptions
    ) -> RenderResult:
        height = console_.height - HEADER_SIZE - BORDER_LINES
        options.height = height
        lines = console_.render_lines(Renderables(self._renderables))

        yield SegmentLines(lines=lines[-height:], new_lines=True)


class HopprJobsPanel(HopprBasePanel):
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress(
        "{task.description}",
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    )

    def __init__(self) -> None:
        super().__init__(renderable=self.progress_bar, title="Jobs")

    def add_task(self, description: str, total: int = 1, **fields) -> TaskID:
        """
        Add a task to be tracked in the `jobs` side panel

        Args:
            description (str): A description of the task

        Returns:
            int: ID of the added task
        """
        return self.progress_bar.add_task(description, total=total, start=False, **fields)


class HopprOutputPanel(HopprBasePanel):
    """
    Customized Rich Text box to simulate a console
    """

    hoppr_console = HopprConsole()

    def __init__(self) -> None:
        super().__init__(renderable=self.hoppr_console)

    def get_width(self) -> int:
        """
        Return the width of the writeable area of the output panel

        Returns:
            int: Width of the output panel
        """
        if hoppr.utils.is_basic_terminal():
            return console.width

        return int(console.width * OUTPUT_PANEL_RATIO) - BORDER_LINES - PAD_LEFT - PAD_RIGHT

    def print(self, msg: RenderableType) -> None:
        """
        Write a message or renderable object to the console output panel

        Args:
            msg (RenderableType): Message or renderable object to write
        """
        self.hoppr_console.append(msg)


class HopprProgressPanel(HopprBasePanel):  # pylint: disable=too-few-public-methods
    """
    Customized Rich Progress bar Panel
    """

    progress_bar = Progress()

    def __init__(self) -> None:
        super().__init__(renderable=self.progress_bar, title="Progress")
        self.progress_bar.add_task(description="All Jobs")
        self.progress_task = self.progress_bar.tasks[0]


class HopprSpinnerColumn(SpinnerColumn):
    """
    Customized SpinnerColumn that displays either red "X" or green check mark based on task result
    """

    _SUCCESS_CHECK: Final[str] = "[green]\u2714"
    _FAIL_X: Final[str] = "\u274c\ufe0f"
    _WARNING_SIGN: Final[str] = "\u26a0\ufe0f"

    @override
    def __init__(  # pylint: disable=too-many-arguments
        self,
        spinner_name: str = "dots",
        style: StyleType | None = "[white]progress.spinner",
        speed: float = 1,
        finished_text: TextType = " ",
        table_column: Column | None = None,
    ):
        super().__init__(spinner_name, style, speed, finished_text, table_column)

    @override
    def render(self, task: Task) -> RenderableType:
        status = task.fields.get("status", "")

        match status.lower():
            case "fail":
                self.finished_text = Text.from_markup(self._FAIL_X)
            case "warn":
                self.finished_text = Text.from_markup(self._WARNING_SIGN)
            case _:
                self.finished_text = Text.from_markup(self._SUCCESS_CHECK)

        return self.finished_text if task.finished else self.spinner.render(task.get_time())


class HopprLayout(Layout):
    """
    Layout of the Hoppr console application
    """

    name: str = "root"
    job_id_map: MutableMapping[str, TaskID] = {}
    jobs_panel = HopprJobsPanel()
    output_panel = HopprOutputPanel()
    overall_progress = HopprProgressPanel()

    def __init__(self, title: str = f"Hoppr v{hoppr.__version__}") -> None:
        super().__init__()

        self.split(Layout(name="header", size=HEADER_SIZE), Layout(name="main"))
        self["main"].split_row(Layout(name="side"), Layout(name="console", ratio=OUTPUT_PANEL_RATIO_SIZE))
        self["side"].split(Layout(name="jobs"), Layout(name="progress", size=3))

        # Initialize header
        header = Text(text=title, style="bold blue", justify="center")
        self["header"].update(renderable=HopprBasePanel(renderable=header))

        # Initialize jobs side bar panel
        self["jobs"].update(renderable=self.jobs_panel)

        # Initialize overall progress side bar
        self["progress"].update(renderable=self.overall_progress)
        self.progress_task = self.overall_progress.progress_bar.tasks[0]
        self.progress_task.total = 0

        # Initialize main body panel
        self["console"].update(renderable=self.output_panel)

        def _showwarning(message: Warning | str, category: type[Warning], *_, **__):
            self.output_panel.print("")
            self.print(msg=str(message), source=f"[bold yellow]{category.__name__}")
            self.output_panel.print("")

        # Override `warnings.showwarning` to print to console
        warnings.showwarning = _showwarning

    def add_job(self, description: str, total: int = 1, **fields) -> TaskID:
        """
        Add a job to the `jobs` side panel

        Args:
            description (str): Description of the job to add
        """
        self.progress_task.total = self.progress_task.total or 0

        self.progress_task.total += total

        if not self.job_id_map.get(description):
            self.job_id_map[description] = self.jobs_panel.add_task(description, total=total, **fields)

        return self.job_id_map[description]

    def advance_job(self, name: str):
        """
        Advance progress of a job in the jobs panel and overall progress

        Args:
            name (str): Name of the job to advance
        """
        if (task_id := self.job_id_map.get(name)) is None:
            return

        self.jobs_panel.progress_bar.advance(task_id)
        self.overall_progress.progress_bar.advance(self.overall_progress.progress_task.id)

    def is_job_finished(self, name: str) -> bool:
        """
        Check if a job is finished

        Args:
            name (str): Name of the job
        """
        if (task_id := self.job_id_map.get(name)) is None:
            return True

        return self.jobs_panel.progress_bar.tasks[task_id].finished

    def print(self, msg: RenderableType, source: str | None = None, style: str = "") -> None:
        """
        Write a message or renderable object to the console output panel

        Args:
            msg (RenderableType): Message or renderable object to write
            source (str | None, optional): Message sender. Defaults to None.
            style (str, optional): Style to apply to message. Defaults to "".
        """
        output_width = self.output_panel.get_width()

        table = Table(
            box=None,
            collapse_padding=True,
            pad_edge=False,
            show_edge=False,
            show_header=False,
            width=output_width,
        )

        msg_column = Column(width=output_width)
        table.columns = [msg_column]
        row_values: list[RenderableType] = []

        if source:
            source_length = len(Text.from_markup(source))
            source_column = Column(width=source_length + PAD_RIGHT)
            table.columns = [source_column, *table.columns]
            msg_column.width = output_width - source_length
            row_values.append(Text.from_markup(source, style="bold cyan"))

        row_values.append(msg)
        table.add_row(*row_values)
        msg = table

        if hoppr.utils.is_basic_terminal():
            console.print(msg, style=style)
        else:
            self.output_panel.print(msg)

    def start_job(self, name: str):
        """
        Start a job in the jobs panel

        Args:
            name (str): Name of the job to start
        """
        if (task_id := self.job_id_map.get(name)) is None:
            return

        self.jobs_panel.progress_bar.start_task(task_id)

    def stop_job(self, name: str):
        """
        Stop a job in the jobs panel

        Args:
            name (str): Name of the job to stop
        """
        if (task_id := self.job_id_map.get(name)) is None:
            return

        self.jobs_panel.progress_bar.stop_task(task_id)

    def update_job(self, name: str, **fields):
        """
        Update a job in the jobs panel

        Args:
            name (str): Name of the job to update
        """
        if (task_id := self.job_id_map.get(name)) is None:
            return

        self.jobs_panel.progress_bar.update(task_id, **fields)



"""
Collector plugin for RPM packages
"""
from __future__ import annotations

import gzip
import time
import warnings

from collections import OrderedDict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, MutableMapping

import jmespath
import requests
import xmltodict

from pydantic import SecretStr
from requests import HTTPError, Response
from requests.auth import HTTPBasicAuth

import hoppr.net
import hoppr.utils

from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_process, hoppr_rerunner
from hoppr.constants import BomProps
from hoppr.exceptions import HopprExperimentalWarning, HopprPluginError
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.models.manifest import SearchSequence
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL


warnings.filterwarnings(action="once", category=HopprExperimentalWarning)
warnings.warn(
    message="The RPM collector plugin is experimental; use at your own risk.",
    category=HopprExperimentalWarning,
    stacklevel=2,
)


class CollectRpmPlugin(SerialCollectorPlugin):
    """
    Collector plugin for RPM packages
    """

    REQUEST_RETRIES: Final[int] = 3
    REQUEST_RETRY_INTERVAL: Final[float] = 5
    REQUEST_TIMEOUT: Final[float] = 60

    products: list[str] = ["rpm/*"]
    required_commands = []
    supported_purl_types: list[str] = ["rpm"]
    system_repositories: list[str] = []

    rpm_data: MutableMapping[str, OrderedDict[str, Any]] = {}

    def get_version(self) -> str:
        return hoppr.__version__

    def _check_artifact_hash(self, dest_file: Path, hash_algorithm: str, hash_string: str):
        computed_hash = hoppr.net.get_file_hash(artifact=dest_file, algorithm=hash_algorithm)

        if computed_hash != hash_string:
            self.get_logger().debug("Computed hash: %s, expected: %s", computed_hash, hash_string, indent_level=2)
            raise HopprPluginError(f"Checksum for {Path(dest_file).name} does not match expected hash.")

    def _download_component(self, download_url: str, dest_file: Path, creds: CredentialRequiredService | None = None):
        self.get_logger().info(msg="Downloading RPM package:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_url}", indent_level=3)
        self.get_logger().info(msg=f"destination: {dest_file}", indent_level=3)

        # Download the RPM file to the target directory
        response = hoppr.net.download_file(  # pylint: disable=duplicate-code
            url=download_url,
            dest=str(dest_file),
            creds=creds,
            proxies=self._get_proxies_for(download_url),
        )

        if not (download_result := Result.from_http_response(response=response)).is_success():
            raise HopprPluginError(download_result.message)

    def _get_download_url(self, purl: PackageURL, repo_url: str) -> tuple[str, str, str]:
        """Get information required to download and verify an RPM package.

        Args:
            purl: PackageURL of component attributes
            repo_url: The RPM repository URL

        Returns:
            Tuple containing download URL, hash algorithm, and checksum string.

        Raises:
            HopprPluginError: malformed/missing `purl.version`, or component not found in repodata
        """
        try:
            rpm_version, rpm_release = purl.version.split("-")
        except (AttributeError, ValueError) as ex:
            raise HopprPluginError(f"Failed to parse version string from PURL: '{purl}'") from ex

        arch = purl.qualifiers.get("arch") or "noarch"

        if (check_result := self.check_purl_specified_url(purl, repo_url)).is_fail():
            if self.context.strict_repos:
                raise HopprPluginError(check_result.message)

            repo_url = purl.qualifiers.get("repository_url", repo_url)
            self._populate_rpm_data(RepositoryUrl(url=repo_url))

        self.get_logger().debug("Searching RPM data for package with attributes:", indent_level=2)
        self.get_logger().debug("name:    %s", purl.name, indent_level=3)
        self.get_logger().debug("version: %s", rpm_version, indent_level=3)
        self.get_logger().debug("release: %s", rpm_release, indent_level=3)
        self.get_logger().debug("arch:    %s", arch, indent_level=3)

        component_data = jmespath.search(
            expression=f"""
                metadata.package[? name=='{purl.name}' &&
                version."@ver"=='{rpm_version}' &&
                version."@rel"=='{rpm_release}' &&
                (arch=='{arch}' || arch=='noarch')] | [0]
            """,
            data=self.rpm_data[repo_url],
        )

        if not component_data:
            raise HopprPluginError(f"RPM package not found in repository: '{purl}'")

        download_url, hash_algorithm, hash_string = jmespath.search(
            expression='[location."@href", checksum."@type", checksum."#text"]',
            data=component_data,
        )

        download_url = f"{repo_url}/{download_url}"

        return download_url, hash_algorithm, hash_string

    def _get_primary_xml_data(
        self,
        repo_url: RepositoryUrl,
        repomd_dict: OrderedDict[str, Any],
        auth: HTTPBasicAuth | None = None,
    ) -> OrderedDict[str, Any]:
        primary_xml_url = jmespath.search(
            expression="""repomd.data[? "@type"=='primary'].location."@href" | [0]""",
            data=repomd_dict,
        )

        primary_xml_url = repo_url / primary_xml_url

        try:
            response = self._stream_url_data(url=primary_xml_url, auth=auth)
        except HTTPError as ex:
            raise HopprPluginError(f"Failed to get primary XML data from {primary_xml_url}") from ex

        data: OrderedDict[str, Any] = OrderedDict()

        # Parse primary XML data to dict
        for chunk in response.iter_content(chunk_size=None):
            data = xmltodict.parse(xml_input=gzip.decompress(chunk), force_list=["package"])

        return data

    def _get_repodata(self, repo_url: RepositoryUrl, auth: HTTPBasicAuth | None = None) -> OrderedDict[str, Any]:
        repomd_url = repo_url / "repodata" / "repomd.xml"

        try:
            response = self._stream_url_data(url=repomd_url, auth=auth)
        except HTTPError as ex:
            raise HopprPluginError(f"Failed to get repository metadata from {repo_url}") from ex

        repomd_dict: OrderedDict[str, Any] = xmltodict.parse(xml_input=response.text, force_list=["data"])

        # Download all metadata files listed in repomd.xml to bundle directory for this repo
        metadata_files = jmespath.search(expression='repomd.data[].location."@href"', data=repomd_dict)
        repodata_dir = self.directory_for(purl_type="rpm", repo_url=str(repo_url), subdir="repodata")

        for metadata_file in metadata_files:
            self.get_logger().debug("Downloading %s", metadata_file, indent_level=2)
            hoppr.net.download_file(url=str(repo_url / metadata_file), dest=str(repodata_dir.parent / metadata_file))

        return repomd_dict

    def _populate_rpm_data(self, repo_url: RepositoryUrl) -> None:
        """
        Populate `rpm_data` dict for a repository

        Args:
            repo_url (str): The RPM repository URL
        """
        if type(self).rpm_data.get(f"{repo_url}"):
            return

        self.get_logger().debug("Populating RPM data for repository: %s", repo_url, indent_level=1)

        auth: HTTPBasicAuth | None = None

        if (creds := Credentials.find(f"{repo_url}")) and isinstance(creds.password, SecretStr):
            auth = HTTPBasicAuth(username=creds.username, password=creds.password.get_secret_value())

        try:
            repomd_dict: OrderedDict[str, Any] = self._get_repodata(repo_url, auth=auth)
            primary_xml_data: OrderedDict[str, Any] = self._get_primary_xml_data(repo_url, repomd_dict, auth)
        except HopprPluginError as ex:
            raise ex

        type(self).rpm_data[f"{repo_url}"] = primary_xml_data

    def _stream_url_data(self, url: RepositoryUrl, auth: HTTPBasicAuth | None = None) -> Response:
        """Stream download data from specified URL.

        Args:
            url: URL of remote resource to stream.
            auth: Basic authentication if required by URL. Defaults to None.

        Raises:
            HTTPError: Failed to download resource after 3 attempts.

        Returns:
            The web request response.
        """
        response = Response()

        for _ in range(self.REQUEST_RETRIES):
            response = requests.get(
                url=f"{url}",
                auth=auth,
                stream=True,
                timeout=self.REQUEST_TIMEOUT,
                proxies=self._get_proxies_for(f"{url}"),
            )

            try:
                response.raise_for_status()
                return response
            except HTTPError:
                time.sleep(self.REQUEST_RETRY_INTERVAL)

        raise HTTPError(f"Failed to retrieve data from {url}", response=response)

    @hoppr_process
    @hoppr_rerunner
    def pre_stage_process(self) -> Result:
        """
        Populate RPM data mapping for repositories
        """
        # Get all repository search sequences from all components
        results: list[str] = jmespath.search(
            expression=f"""
                components[*] | [? starts_with(purl, 'pkg:rpm')].properties[] |
                [? name=='{BomProps.COMPONENT_SEARCH_SEQUENCE.value}'].value
            """,
            data=self.context.consolidated_sbom.dict(),
        )

        # Parse and flatten repositories from component search sequence JSON strings, then remove duplicates
        search_repos = [repo for result in results for repo in SearchSequence.parse_raw(result).repositories]
        search_repos = hoppr.utils.dedup_list(search_repos)

        for repo_url in [RepositoryUrl(url=url) for url in search_repos]:
            self._populate_rpm_data(repo_url)

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        purl = hoppr.utils.get_package_url(comp.purl)

        try:
            download_url, hash_algorithm, hash_string = self._get_download_url(purl=purl, repo_url=repo_url)
        except HopprPluginError as ex:
            return Result.retry(message=str(ex))

        target_dir = self.directory_for(
            purl_type="rpm",
            repo_url=repo_url,
            subdir=Path(download_url).parent.relative_to(repo_url),
        )

        dest_file = target_dir / Path(download_url).name

        try:
            self._download_component(download_url, dest_file, creds)
            self._check_artifact_hash(dest_file, hash_algorithm, hash_string)
        except HopprPluginError as ex:
            return Result.fail(message=str(ex))

        self.set_collection_params(comp, repo_url, target_dir)

        return Result.success(return_obj=comp)



"""
Transfer file data model
"""
from __future__ import annotations

import math

from enum import Enum
from importlib.metadata import entry_points
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Literal

from pydantic import Field, root_validator, validator
from typing_extensions import Self

import hoppr.utils

from hoppr.models.base import HopprBaseModel, HopprBaseSchemaModel

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny
else:
    DictStrAny = dict[str, Any]

_DELTA_SBOM_REFS = {"DeltaSbom", "delta_sbom", "hoppr.core_plugins.delta_sbom"}

# Constrained string type for stage key name
StageName = Annotated[str, Field(regex=r"^\S+", min_length=1)]


class Plugin(HopprBaseModel):
    """
    Plugin data model
    """

    name: str = Field(..., description="Name of plugin")
    config: DictStrAny | None = Field(None, description="Mapping of additional plugin configuration settings to values")

    @root_validator(pre=True)
    @classmethod
    def validate_plugin(cls, values: DictStrAny):
        """
        Validate Plugin model
        """
        name = values.get("name")
        plugin_eps = entry_points(group="hoppr.plugin")

        for plugin in plugin_eps:
            # Check project entry points for plugin name, module, or class
            if str(name) in {plugin.name, plugin.module, plugin.attr}:
                values["name"] = plugin.module
                break

        return values


Plugin.update_forward_refs()


class ComponentCoverage(Enum):
    """
    Enumeration to indicate how often each component should be processed
    """

    OPTIONAL = (0, math.inf)
    EXACTLY_ONCE = (1, 1)
    AT_LEAST_ONCE = (1, math.inf)
    NO_MORE_THAN_ONCE = (0, 1)

    def __init__(self, min_allowed: int, max_allowed: int):
        self.min_value = min_allowed
        self.max_value = max_allowed

    def __str__(self) -> str:
        return str(self.name)

    def accepts_count(self, count: int) -> bool:
        """
        Identifies whether a specified count is acceptable for this coverage value
        """
        return self.min_value <= count <= self.max_value


class Stage(HopprBaseModel):
    """
    Stage data model
    """

    component_coverage: Literal["AT_LEAST_ONCE", "EXACTLY_ONCE", "NO_MORE_THAN_ONCE", "OPTIONAL"] | None = Field(
        default=None, exclude=True, description="Defines how often components should be processed"
    )
    plugins: list[Plugin] = Field(..., description="List of Hoppr plugins to load")


class StageRef(Stage):
    """
    StageRef data model
    """

    name: StageName


Stages = Annotated[dict[StageName, Stage], ...]


class TransferFile(HopprBaseSchemaModel):
    """
    Transfer file data model
    """

    kind: Literal["Transfer"]
    max_processes: int | None = Field(3, description="Max processes to create when running Hoppr application")
    stages: Stages = Field(..., description="Mapping of stage names to property definitions")

    @classmethod
    def parse_file(cls, path: str | Path, *args, **kwargs) -> Self:  # pylint: disable=unused-argument
        """
        Override to resolve local file paths relative to transfer file
        """
        path = Path(path)

        data = hoppr.utils.load_file(path)
        if not isinstance(data, dict):
            raise TypeError("Local file content was not loaded as dictionary")

        stages = data.get("stages", {})
        plugins = [plugin for stage in stages.values() for plugin in stage.get("plugins", [])]

        # Resolve DeltaSbom previous delivery path relative to transfer file
        for plugin in plugins:
            if plugin.get("name") in _DELTA_SBOM_REFS and plugin.get("config", {}).get("previous"):
                previous = (path.parent / plugin["config"]["previous"]).resolve()
                plugin["config"]["previous"] = str(previous)

        return cls(**data)


class Transfer(TransferFile):
    """
    Transfer data model
    """

    stages: list[StageRef]  # type: ignore[assignment]

    @validator("stages", allow_reuse=True, pre=True)
    @classmethod
    def validate_stages(cls, stages: DictStrAny) -> list[StageRef]:
        """
        Transform Stages into list of StageRef objects
        """
        stage_refs: list[StageRef] = []
        plugin_names = set()

        for stage_name, stage in stages.items():
            stage["name"] = stage_name
            stage_refs.append(StageRef.parse_obj(stage))

            plugin_names.update({plugin["name"] for plugin in stage["plugins"]})

        if not plugin_names.intersection(_DELTA_SBOM_REFS):
            stage_refs = [
                StageRef(name=StageName("_delta_sbom_"), plugins=[Plugin(name="delta_sbom", config=None)]),
                *stage_refs,
            ]

        return stage_refs

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Self:
        """
        Load transfer file from local path or dict
        """
        match source:
            case dict():
                return cls.parse_obj(source)
            case str() | Path():
                return cls.parse_file(source)
            case _:
                raise TypeError("'source' argument must be one of: 'str', 'Path', 'dict[str, Any]'")

    def yaml(self, *args, **kwargs) -> str:
        transfer_dict = self.dict(by_alias=True)
        transfer_dict["stages"] = {str(stage.name): {"plugins": stage.plugins} for stage in self.stages}
        return TransferFile.parse_obj(transfer_dict).yaml(*args, **kwargs)



"""
Manifest file data model
"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Annotated, ClassVar, Literal, MutableMapping

from pydantic import Field, NoneStr, create_model, validator
from pydantic.main import ModelMetaclass
from requests import HTTPError
from typer import secho

import hoppr.net
import hoppr.oci_artifacts
import hoppr.utils

from hoppr.constants import BomProps
from hoppr.exceptions import HopprLoadDataError
from hoppr.models.base import HopprBaseModel, HopprBaseSchemaModel
from hoppr.models.sbom import Component, ExternalReference, Property, Sbom, SbomRef
from hoppr.models.types import LocalFile, OciFile, PurlType, RepositoryUrl, UrlFile

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class Repository(HopprBaseModel):
    """
    Repository data model
    """

    url: str = Field(
        ...,
        regex=r"^([^:/?#]+:(?=//))?(//)?(([^:]+(?::[^@]+?)?@)?[^@/?#:]*(?::\d+?)?)?[^?#]*(\?[^#]*)?(#.*)?",
    )
    description: NoneStr = None

    @validator("url", pre=True)
    @classmethod
    def validate_url_pre(cls, url: str | dict[str, str]) -> str:
        """
        Ensure URL is parsed as `str` before Pydantic validation
        """
        if isinstance(url, dict):
            url = url["url"]

        return str(url)

    @validator("url")
    @classmethod
    def validate_url(cls, url: str) -> RepositoryUrl:
        """
        Convert URL string to `RepositoryUrl` after validation
        """
        return RepositoryUrl(url=url)


# Dynamic Repositories model with PurlType values as attribute names
purl_type_repo_mapping = {str(purl_type): (list[Repository], Field([], unique_items=True)) for purl_type in PurlType}
RepositoriesMetaclass: ModelMetaclass = create_model(  # type: ignore[call-overload]
    "RepositoriesMetaclass", __base__=HopprBaseModel, **purl_type_repo_mapping
)


class Repositories(RepositoriesMetaclass):  # type: ignore[misc, valid-type]
    """
    Repositories data model
    """

    # pylint: disable=too-few-public-methods

    def __getitem__(self, item: str | PurlType) -> list[Repository]:
        return getattr(self, str(item), [])

    def __setitem__(self, item: str | PurlType, value: list[Repository]) -> None:
        setattr(self, str(item), value)

    def __iter__(self):
        for purl_type in PurlType:
            yield (purl_type, self[purl_type])


class SearchSequence(HopprBaseModel):
    """
    SearchSequence data model
    """

    version: Literal["v1"] = "v1"
    repositories: list[RepositoryUrl | str] = []


IncludeRef = Annotated[LocalFile | UrlFile, Field(..., description="Reference to a local or remote manifest file")]
Includes = Annotated[list[IncludeRef], Field(..., description="List of manifest files to load")]
IncludeRefMap = Annotated[MutableMapping[IncludeRef, "ManifestFile"], Field(...)]


class ManifestFile(HopprBaseSchemaModel):
    """
    Data model to describe a single manifest file
    """

    kind: Literal["Manifest"]
    sboms: list[SbomRef] = []
    repositories: Repositories = Field(..., description="Maps supported PURL types to package repositories/registries")
    includes: Includes = []

    @classmethod
    def parse_file(cls, path: str | Path, *args, **kwargs) -> ManifestFile:  # pylint: disable=unused-argument
        """
        Override to resolve local file paths relative to manifest file
        """
        path = Path(path)

        data = hoppr.utils.load_file(path)
        if not isinstance(data, dict):
            raise TypeError("Local file content was not loaded as dictionary")

        # Resolve local file path references relative to manifest file path
        for sbom in data.get("sboms", []):
            if "local" in sbom:
                local_ref = (path.parent / sbom["local"]).resolve()
                local_ref = local_ref.relative_to(Path.cwd()) if local_ref.is_relative_to(Path.cwd()) else local_ref

                sbom["local"] = str(local_ref)

        for include in data.get("includes", []):
            if "local" in include:
                local_ref = (path.parent / include["local"]).resolve().relative_to(Path.cwd())
                include["local"] = str(local_ref)

        return cls(**data)

    @classmethod
    def parse_obj(cls, obj: DictStrAny) -> ManifestFile:
        """
        Override to remove local file paths that can't be resolved
        """
        for include in list(obj.get("includes", [])):
            if "local" in include and not Path(include["local"]).is_absolute():
                secho(f"Skipping local include: relative path '{include['local']}' cannot be resolved", fg="yellow")
                obj["includes"].remove(include)

        for sbom in list(obj.get("sboms", [])):
            if "local" in sbom and not Path(sbom["local"]).is_absolute():
                secho(f"Skipping local SBOM: relative path '{sbom['local']}' cannot be resolved", fg="yellow")
                obj["sboms"].remove(sbom)

        return cls(**obj)


class Manifest(ManifestFile):
    """
    Manifest data model that generates lookups for `includes` and `sboms` references
    """

    consolidated_sbom: Annotated[Sbom, Field(exclude=True)] = None  # type: ignore[assignment]

    # Attributes not included in schema
    loaded_manifests: ClassVar[IncludeRefMap] = {}

    @validator("includes", allow_reuse=True, always=True)
    @classmethod
    def populate_loaded_manifests(cls, includes: Includes, values: DictStrAny) -> Includes:
        """
        Validator that automatically loads manifest from local file or URL into lookup dictionary
        """
        include_refs = [ref for ref in includes if ref not in cls.loaded_manifests]

        for include_ref in include_refs:
            manifest_file = cls._load_include(include_ref)

            # Prepend all repositories of current manifest to included manifest
            for purl_type, repos in values.get("repositories", Repositories()):
                include_repos = [repo.copy(deep=True) for repo in manifest_file.repositories[str(purl_type)]]
                manifest_file.repositories[str(purl_type)] = hoppr.utils.dedup_list([*repos, *include_repos])

            cls.loaded_manifests[include_ref] = manifest_file
            cls.load(manifest_file.dict())

        return includes

    @validator("sboms", allow_reuse=True, always=True)
    @classmethod
    def populate_loaded_sboms(cls, sboms: list[SbomRef]) -> list[SbomRef]:
        """
        Validator that automatically loads SBOM from local file or URL into lookup dictionary
        """
        sbom_refs = [ref for ref in sboms if ref not in Sbom.loaded_sboms]

        for sbom_ref in sbom_refs:
            cls._load_sbom(sbom_ref)

        return sboms

    @validator("repositories", always=True)
    @classmethod
    def populate_repositories(cls, repositories: Repositories, values: DictStrAny) -> Repositories:
        """
        Populate SBOM components with a property representing repositories from manifest
        """
        sbom_refs = values.get("sboms", [])

        for sbom_ref in sbom_refs:
            loaded_sbom = cls._load_sbom(sbom_ref)

            # Merge current SBOM metadata into consolidated SBOM
            for component in [comp for comp in loaded_sbom.components if comp.purl]:
                purl_type = hoppr.utils.get_package_url(component.purl).type
                manifest_repos = [str(repo.url) for repo in repositories[purl_type]]
                cls._add_repository_search_sequence(component, manifest_repos)

                # Add external reference to SBOM file that includes this component
                component.externalReferences.append(
                    ExternalReference(
                        url=str(sbom_ref),
                        type="bom",  # type: ignore[arg-type]
                        comment=loaded_sbom.serialNumber,
                        hashes=None,
                    )
                )

                # Merge component into consolidated SBOM
                cls._add_component(component)

        return repositories

    @validator("consolidated_sbom", allow_reuse=True, always=True)
    @classmethod
    def populate_consolidated_sbom(cls, consolidated_sbom: Sbom | None) -> Sbom:
        """
        Populate `consolidated_sbom` with previously loaded `components` and `externalReferences`
        """
        consolidated_sbom = Sbom()

        for sbom_ref, sbom in Sbom.loaded_sboms.items():
            external_ref = ExternalReference(
                url=str(sbom_ref),
                type="bom",  # type: ignore[arg-type]
                comment=sbom.serialNumber,
            )

            sbom.externalReferences = hoppr.utils.dedup_list([*sbom.externalReferences, external_ref])

            for component in sbom.components:
                component.externalReferences = hoppr.utils.dedup_list([*component.externalReferences, external_ref])

            consolidated_sbom.merge(sbom)

        return consolidated_sbom

    @classmethod
    def _add_component(cls, component: Component) -> None:
        # Merge component into previously loaded component
        if loaded := Component.find(component.bom_ref):
            loaded.merge(component)
        else:
            Component.unique_id_map[component.bom_ref] = component

    @classmethod
    def _add_repository_search_sequence(cls, component: Component, manifest_repos: list[str]) -> None:
        # Create default search sequence
        search_sequence = SearchSequence()

        # Try to get existing repository search sequence property
        search_sequence_props = [
            property_.copy(deep=True)
            for property_ in component.properties
            if property_.name == BomProps.COMPONENT_SEARCH_SEQUENCE
        ]

        for property_ in search_sequence_props:
            search_repos = SearchSequence.parse_raw(property_.value or "{}").repositories
            search_sequence.repositories.extend(search_repos)
            component.properties.remove(property_)

        # Generate the repository search sequence
        search_sequence.repositories.extend(manifest_repos)
        search_sequence.repositories = hoppr.utils.dedup_list(search_sequence.repositories)

        # Add repository search sequence as component property
        component.properties.append(
            Property(
                name=BomProps.COMPONENT_SEARCH_SEQUENCE,
                value=search_sequence.json(),
            )
        )

    @classmethod
    def _load_include(cls, include_ref: IncludeRef) -> ManifestFile:
        """Load manifest include from local or URL reference or return manifest that was previously loaded."""
        loaded: ManifestFile | None = None

        match include_ref:
            case LocalFile():
                loaded = cls.find("local", location=include_ref.local) or ManifestFile.parse_file(include_ref.local)
            case UrlFile():
                if not (loaded := cls.find("url", location=include_ref.url)):
                    include_dict = hoppr.net.load_url(include_ref.url)
                    if not isinstance(include_dict, dict):
                        raise TypeError("URL manifest include was not loaded as dictionary")

                    loaded = ManifestFile.parse_obj(include_dict)

        return loaded

    @classmethod
    def _load_sbom(cls, sbom_ref: SbomRef) -> Sbom:
        """Load SBOM from local or URL reference or return SBOM with that reference that was previously loaded."""
        loaded: Sbom | None = None

        match sbom_ref:
            case LocalFile():
                loaded = Sbom.find_ref(ref_type="local", location=sbom_ref.local) or Sbom.load(sbom_ref.local)
            case OciFile():
                if not (loaded := Sbom.find_ref(ref_type="oci", location=sbom_ref.oci)):
                    data = hoppr.oci_artifacts.pull_artifact(sbom_ref.oci)
                    if not isinstance(data, dict):
                        raise TypeError("OCI URL SBOM file was not loaded as dictionary")

                    loaded = Sbom.load(data)
            case UrlFile():
                loaded = Sbom.find_ref(ref_type="url", location=sbom_ref.url) or Sbom.load(sbom_ref.url)

        return loaded

    @classmethod
    def find(cls, ref_type: Literal["local", "url"], location: str | Path) -> ManifestFile | None:
        """
        Lookup manifest object by include reference

        Args:
            ref_type (Literal["local", "url"]): Type of include
            location (str | Path): Path to included manifest

        Returns:
            ManifestFile | None: Manifest object if found, otherwise None
        """
        match ref_type:
            case "local":
                location = Path(location).resolve().relative_to(Path.cwd())
                return cls.loaded_manifests.get(LocalFile(local=Path(location)))
            case "url":
                return cls.loaded_manifests.get(UrlFile(url=str(location)))
            case _:
                return None

    @classmethod
    def load(cls, source: str | Path | DictStrAny) -> Manifest:
        """
        Load manifest from local file, URL, or dict
        """
        match source:
            case dict():
                data = source
            case Path():
                # Convert source to relative path if in current working directory subpath
                source = source.resolve()
                source = source.relative_to(Path.cwd()) if source.is_relative_to(Path.cwd()) else source

                manifest_file = ManifestFile.parse_file(source)
                cls.loaded_manifests[LocalFile(local=source)] = manifest_file
                data = manifest_file.dict(by_alias=True)
            case str():
                try:
                    include_dict = hoppr.net.load_url(source)
                    if not isinstance(include_dict, dict):
                        raise TypeError("URL manifest include was not loaded as dictionary")

                    manifest_file = ManifestFile.parse_obj(include_dict)
                    url_ref = UrlFile(url=source)
                    cls.loaded_manifests[url_ref] = manifest_file
                except (HopprLoadDataError, HTTPError) as ex:
                    raise HopprLoadDataError from ex

                data = manifest_file.dict(by_alias=True)

        return cls(**data)



"""
Collector plugin for helm charts
"""
from __future__ import annotations

from pathlib import Path

from pydantic import SecretStr

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.exceptions import HopprLoadDataError
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectHelmPlugin(SerialCollectorPlugin):
    """
    Class to copy helm charts
    """

    supported_purl_types = ["helm"]
    required_commands = ["helm"]
    products: list[str] = ["helm/*"]
    system_repositories: list[str] = []

    def get_version(self) -> str:  # pylint: disable=duplicate-code
        return __version__

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

        self.required_commands = (self.config or {}).get("helm_command", self.required_commands)
        self.base_command = [self.required_commands[0], "fetch"]

        system_repos_file = Path.home() / ".config" / "helm" / "repositories.yaml"
        if not self.context.strict_repos and system_repos_file.exists():
            system_repos: list[dict[str, str]] = []

            try:
                system_repos_dict = hoppr.utils.load_file(input_file_path=system_repos_file)
                if not isinstance(system_repos_dict, dict):
                    raise HopprLoadDataError("Incorrect format.")

                system_repos = system_repos_dict["repositories"]
            except HopprLoadDataError as ex:
                self.get_logger().warning(msg=f"Unable to parse Helm repositories file ({system_repos_file}): '{ex}'")

            self.system_repositories.extend(repo["url"] for repo in system_repos)

    @hoppr_rerunner
    # pylint: disable=duplicate-code
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None):
        """
        Collect helm chart
        """

        purl = hoppr.utils.get_package_url(comp.purl)

        target_dir = self.directory_for(purl.type, repo_url, subdir=f"{purl.name}_{purl.version}")

        run_result = None
        for subdir in ["", purl.name]:
            source_url = RepositoryUrl(url=repo_url) / subdir

            self.get_logger().info(msg="Fetching helm chart:", indent_level=2)
            self.get_logger().info(msg=f"source: {source_url}", indent_level=3)
            self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

            command = [
                *self.base_command,
                "--repo",
                f"{source_url}",
                "--destination",
                f"{target_dir}",
                purl.name,
                "--version",
                purl.version,
                *(["--debug"] if self.get_logger().is_verbose() else []),
            ]

            password_list = []

            if creds is not None and isinstance(creds.password, SecretStr):
                command = [
                    *command,
                    "--username",
                    creds.username,
                    "--password",
                    creds.password.get_secret_value(),
                ]

                password_list = [creds.password.get_secret_value()]

            run_result = self.run_command(command, password_list)

            if run_result.returncode == 0:
                self.get_logger().info("Complete helm chart artifact copy for %s version %s", purl.name, purl.version)
                self.set_collection_params(comp, repo_url, target_dir)

                return Result.success(return_obj=comp)

        msg = f"Failed to download {purl.name} version {purl.version} helm chart from {repo_url}."

        if run_result is not None and "404 Not Found" in str(run_result.stderr):
            self.get_logger().debug(msg=msg, indent_level=2)
            return Result.fail(f"{msg} Chart not found.")

        self.get_logger().debug(msg=msg, indent_level=2)
        return Result.retry(msg)



"""
Top-level CLI application
"""
from __future__ import annotations

import ctypes
import sys

from pathlib import Path
from platform import python_version

import rich

from rich.ansi import AnsiDecoder
from typer import Typer

import hoppr.utils

from hoppr.cli import bundle, generate, merge, validate


# Windows flags and types
NT_ENABLE_ECHO_INPUT = 0b0100
NT_ENABLE_LINE_INPUT = 0b0010
NT_ENABLE_PROCESSED_INPUT = 0b0001
NT_CONSOLE_FLAGS = NT_ENABLE_ECHO_INPUT | NT_ENABLE_LINE_INPUT | NT_ENABLE_PROCESSED_INPUT
NT_STD_OUTPUT_HANDLE = ctypes.c_uint(-11)

# Enable ANSI processing on Windows systems
if sys.platform == "win32":  # pragma: no cover
    nt_kernel = ctypes.WinDLL(name="kernel32.dll")

    nt_kernel.SetConsoleMode(nt_kernel.GetStdHandle(NT_STD_OUTPUT_HANDLE), NT_CONSOLE_FLAGS)


app = Typer(
    name="hopctl",
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Collect, process, & bundle your software supply chain",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
)

app.add_typer(typer_instance=bundle.app, name="bundle")
app.add_typer(typer_instance=generate.app, name="generate")
app.add_typer(typer_instance=merge.app, name="merge")
app.add_typer(typer_instance=validate.app, name="validate")


# Aliases for deprecated commands to preserve backward compatibility
generate_layout = app.command(
    name="generate-layout",
    deprecated=True,
    rich_help_panel="Deprecated",
    help="See `hopctl generate layout` subcommand",
)(generate.layout)


generate_schemas = app.command(
    name="generate-schemas",
    deprecated=True,
    rich_help_panel="Deprecated",
    help="See `hopctl generate schemas` subcommand",
)(generate.schemas)


@app.command()
def version():
    """
    Print version information for `hoppr`
    """
    # Non-TTY terminals and MacOS Terminal.app don't support ANSI multibyte characters. Print low-resolution art instead
    suffix = ".ascii" if hoppr.utils.is_basic_terminal() else ".ansi"
    hippo_file = (Path(hoppr.__file__).parent / "resources" / "hoppr-hippo").with_suffix(suffix)

    decoder = AnsiDecoder()
    hippo = decoder.decode(hippo_file.read_text(encoding="utf-8"))

    rich.print(*hippo, sep="\n")
    rich.print(f"[green]Hoppr Framework Version[/] : {hoppr.__version__}")
    rich.print(f"[green]Python Version         [/] : {python_version()}")


__all__ = ["app"]



"""
Common options for typer commands
"""

from __future__ import annotations

import logging
import os
import textwrap
import time

from pathlib import Path

from pydantic import ValidationError
from typer import BadParameter, CallbackParam, Context, Option

import hoppr.utils

from hoppr.logger import HopprLogger
from hoppr.models.manifest import Manifest
from hoppr.models.sbom import Sbom


_logger: HopprLogger | None = None


def log_file_callback(log_file: Path | None) -> Path:
    """
    Generate a default log file if none was specified

    Args:
        log_file: Path to the log file if specified, otherwise None

    Returns:
        Path to either the log file that was passed or an automatically generated one
    """
    return log_file or Path(f"hoppr_{time.strftime('%Y%m%d-%H%M%S')}.log")


def set_global_option(ctx: Context, param: CallbackParam, value: bool) -> bool:
    """
    Add option value to shared Click context. Allows interspersed CLI options/arguments

    Args:
        ctx: Click context to update
        param: Typer metadata for the parameter
        value: The parameter value

    Returns:
        bool: The flag that was passed, unmodified
    """
    if param.name:
        ctx.params[param.name] = value

        if param.envvar and value:
            os.environ[str(param.envvar)] = "1"

    return value


basic_term_option: bool = Option(
    False,
    "-b",
    "--basic-term",
    callback=set_global_option,
    help="Use simplified output for non-TTY or legacy terminal emulators",
    is_eager=True,
    envvar="HOPPR_BASIC_TERM",
    rich_help_panel="Global options",
    show_default=False,
)

experimental_option: bool = Option(
    False,
    "-x",
    "--experimental",
    callback=set_global_option,
    help="Enable experimental features",
    is_eager=True,
    envvar="HOPPR_EXPERIMENTAL",
    rich_help_panel="Global options",
    show_default=False,
)

log_file_option: Path = Option(
    None,
    "-l",
    "--log",
    callback=log_file_callback,
    help="File to which log will be written",
    is_eager=True,
    envvar="HOPPR_LOG_FILE",
    rich_help_panel="Global options",
    show_default=False,
)

strict_repos_option: bool = Option(
    True,
    "--strict/--no-strict",
    callback=set_global_option,
    help="Utilize only manifest repositories for package collection",
    is_eager=True,
    envvar="HOPPR_STRICT_REPOS",
    rich_help_panel="Global options",
    show_default=False,
)

verbose_option: bool = Option(
    False,
    "-v",
    "--debug",
    "--verbose",
    callback=set_global_option,
    help="Enable debug output",
    is_eager=True,
    rich_help_panel="Global options",
    show_default=False,
)


# ------------------------------------------------------------------------------------------------- #
# SBOM input options
# ------------------------------------------------------------------------------------------------- #
def manifest_callback(manifest_file: Path | None) -> Path | None:
    """
    Load SBOM files from Hoppr manifest and populate the `--sbom-files` parameter variable

    Args:
        manifest_file: Path to manifest file if specified, otherwise None

    Returns:
        The manifest file that was passed, unmodified
    """
    if manifest_file is None:
        return None

    # Load manifest file to populate `Sbom.loaded_sboms`
    Manifest.load(manifest_file)

    return manifest_file


def sbom_callback(ctx: Context, sbom_sources: list[str | Path]) -> list[str | Path]:
    """
    Load SBOM input files

    Args:
        ctx: Click context
        sbom_sources: Local SBOM files or URLs to parse

    Returns:
        The list of directories that was passed, unmodified
    """
    sbom_sources.extend(ctx.params.get("sbom_files", []))

    # Combine SBOM files from all CLI input file arguments
    log_file = Path(ctx.params.get("log_file", ""))
    verbose = bool(ctx.params.get("verbose"))
    _load_sbom_files(sbom_sources, log_file, verbose)

    return sbom_sources


def sbom_dir_callback(ctx: Context, sbom_dirs: list[Path]) -> list[Path]:
    """
    Load SBOM files from input directories and populate the `--sbom-files` parameter variable

    Args:
        ctx: Click context
        sbom_dirs: List of directories containing SBOM files

    Returns:
        The list of directories that was passed, unmodified
    """
    log_file = Path(ctx.params.get("log_file", ""))
    verbose = bool(ctx.params.get("verbose"))

    sbom_files: list[str | Path] = ctx.params.setdefault("sbom_files", [])

    for sbom_dir in sbom_dirs:
        sbom_dir_files: list[str | Path] = list(sbom_dir.glob("*.json"))
        _load_sbom_files(sbom_dir_files, log_file, verbose)
        sbom_files.extend(sbom_dir_files)

    sbom_files = hoppr.utils.dedup_list(sbom_files)

    return sbom_dirs


def _load_sbom_files(sbom_files: list[str | Path], log_file: Path, verbose: bool) -> None:
    """
    Load SBOM input files
    """
    global _logger  # pylint: disable=global-statement
    _logger = _logger or HopprLogger(
        name="hoppr",
        filename=str(log_file),
        level=logging.DEBUG if verbose else logging.INFO,
    )

    for sbom_file in sbom_files:
        try:
            Sbom.load(sbom_file)
        except ValidationError as ex:
            msg = f"'{sbom_file}' failed to validate against the CycloneDX schema"

            _logger.error(f"{msg}:\n{textwrap.indent(text=str(ex), prefix='    ')}")
            raise BadParameter(f"{msg}. See {log_file} for details.") from ex


manifest_file_option: Path = Option(
    None,
    "-m",
    "--manifest",
    callback=manifest_callback,
    dir_okay=False,
    exists=True,
    help="Manifest file containing input SBOMs",
    resolve_path=True,
    show_default=False,
)

sbom_files_option: list[Path] = Option(
    [],
    "-s",
    "--sbom",
    callback=sbom_callback,
    dir_okay=False,
    exists=True,
    help="Path to SBOM file (can be specified multiple times)",
    show_default=False,
)

sbom_dirs_option: list[Path] = Option(
    [],
    "-d",
    "--sbom-dir",
    callback=sbom_dir_callback,
    exists=True,
    file_okay=False,
    help="Directory containing SBOM files (can be specified multiple times)",
    show_default=False,
)

sbom_urls_option: list[str] = Option(
    [],
    "-u",
    "--sbom-url",
    help="SBOM file URL (can be specified multiple times)",
    metavar="URL",
    show_default=False,
)



"""Hoppr exceptions and warnings."""


class HopprError(RuntimeError):
    """Base exception raised within the hoppr app."""


class HopprPluginError(HopprError):
    """Exception raised for errors working with plug-ins."""


class HopprLoadDataError(HopprError):
    """Exception raised for errors loading json/yml data."""


class HopprCredentialsError(HopprError):
    """Exception raised for errors loading credential data."""


class HopprExperimentalWarning(UserWarning):
    """Warning raised when experimental features are accessed."""



"""
Models to express `licenses` fields
"""
from __future__ import annotations

from typing import Annotated, ClassVar, TypeAlias

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from hoppr_cyclonedx_models import spdx
from pydantic import Extra, Field

from hoppr.models.base import CycloneDXBaseModel, UniqueIDMap


class LicensingOrganizationRequired(CycloneDXBaseModel):
    """
    Licensing data model derived from CycloneDXBaseModel with required `organization` field
    """

    organization: cdx.OrganizationalEntity


class LicensingIndividualRequired(CycloneDXBaseModel):
    """
    Licensing data model derived from CycloneDXBaseModel with required `organization` field
    """

    individual: cdx.OrganizationalContact


OrganizationOrIndividualData: TypeAlias = LicensingOrganizationRequired | LicensingIndividualRequired | None


class Licensing(CycloneDXBaseModel, cdx.Licensing):
    """
    Licensing data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for Licensing model"
        extra = Extra.allow

    licensor: Annotated[
        OrganizationOrIndividualData,
        Field(
            description="The individual or organization that grants a license to another individual or organization",
            title="Licensor",
        ),
    ] = None

    licensee: Annotated[
        OrganizationOrIndividualData,
        Field(description="The individual or organization for which a license was granted to", title="Licensee"),
    ] = None

    purchaser: Annotated[
        OrganizationOrIndividualData,
        Field(description="The individual or organization that purchased the license", title="Purchaser"),
    ] = None

    unique_id_map: ClassVar[UniqueIDMap] = {}


class License(CycloneDXBaseModel, cdx.License):
    """
    License data model derived from CycloneDXBaseModel
    """

    class Config(CycloneDXBaseModel.Config):  # pylint: disable=too-few-public-methods
        "Config for License model"

    licensing: Annotated[
        Licensing | None,
        Field(
            description=(
                "Licensing details describing the licensor/licensee, license type, "
                "renewal and expiration dates, and other important metadata"
            ),
            title="Licensing information",
        ),
    ] = None

    def __hash__(self) -> int:  # pragma: no cover
        return hash(self.bom_ref or self.id or self.name or repr(self))

    unique_id_map: ClassVar[UniqueIDMap] = {}


class SPDXLicense(License):
    """
    License data model derived from CycloneDXBaseModel with required `id` field
    """

    # Override to make `id` field required by removing default value
    id: Annotated[
        spdx.LicenseID,
        Field(
            description="A valid SPDX license ID",
            examples=["Apache-2.0"],
            title="License ID (SPDX)",
        ),
    ]

    def __hash__(self) -> int:  # pragma: no cover
        return hash(self.id)

    unique_id_map: ClassVar[UniqueIDMap] = {}


class NamedLicense(License):
    """
    License data model derived from CycloneDXBaseModel with required `name` field
    """

    # Override to make `name` field required by removing default value
    name: Annotated[
        str,
        Field(
            description="If SPDX does not define the license used, this field may be used to provide the license name",
            examples=["Acme Software License"],
            title="License Name",
        ),
    ]

    def __hash__(self) -> int:  # pragma: no cover
        return hash(self.name)

    unique_id_map: ClassVar[UniqueIDMap] = {}


class LicenseMultipleItem(CycloneDXBaseModel):  # pylint: disable=missing-class-docstring
    license: SPDXLicense | NamedLicense

    def __hash__(self) -> int:
        return hash(self.license)

    unique_id_map: ClassVar[UniqueIDMap] = {}


MultipleLicenses: TypeAlias = Annotated[
    list[LicenseMultipleItem],
    Field(
        description="A list of SPDX licenses and/or named licenses.",
        title="Multiple licenses",
    ),
]


class LicenseExpressionItem(CycloneDXBaseModel):
    """
    SPDX license expression data model
    """

    expression: Annotated[
        str,
        Field(
            title="SPDX License Expression",
            examples=["Apache-2.0 AND (MIT OR GPL-2.0-only)", "GPL-3.0-only WITH Classpath-exception-2.0"],
        ),
    ]
    bom_ref: Annotated[
        str | None,
        Field(
            title="BOM Reference",
            description=(
                "An optional identifier which can be used to reference the license elsewhere in the BOM. "
                "Every bom-ref MUST be unique within the BOM."
            ),
            alias="bom-ref",
            min_length=1,
        ),
    ] = None

    unique_id_map: ClassVar[UniqueIDMap] = {}


LicenseExpression: TypeAlias = Annotated[
    list[LicenseExpressionItem],
    Field(
        title="SPDX License Expression",
        description="A tuple of exactly one SPDX License Expression.",
        min_items=1,
        max_items=1,
    ),
]


LicenseChoice: TypeAlias = Annotated[
    MultipleLicenses | LicenseExpression | None,
    Field(
        description="EITHER (list of SPDX licenses and/or named licenses) OR (tuple of one SPDX License Expression)",
        title="License Choice",
    ),
]



"""
Credentials file data model
"""
from __future__ import annotations

import os

from pathlib import Path
from typing import TYPE_CHECKING, Annotated, ClassVar, Literal

from pydantic import Field, FilePath, NoneStr, SecretStr, root_validator

from hoppr.models.base import HopprBaseModel, HopprBaseSchemaModel

if TYPE_CHECKING:
    from pydantic.typing import DictStrAny


class CredentialRequiredService(HopprBaseModel):
    """
    CredentialRequiredService data model
    """

    url: str
    user_env: NoneStr = Field(None)
    pass_env: NoneStr = Field(...)
    username: str = Field(..., alias="user")
    password: NoneStr | SecretStr = Field(None, exclude=True)

    @root_validator(pre=True, allow_reuse=True)
    @classmethod
    def validate_credential_required_service(cls, values: DictStrAny) -> DictStrAny:
        """
        Dynamically parse credentials using specified environment variables
        """
        try:
            if (user_env := values.get("user_env")) is not None:
                if username := os.environ[user_env]:
                    values["user"] = username
                else:
                    raise ValueError(user_env)

            if (pass_env := values.get("pass_env")) is not None:
                if password := os.environ[pass_env]:
                    values["password"] = SecretStr(password)
                else:
                    raise ValueError(pass_env)
        except (KeyError, ValueError) as ex:
            raise ValueError(f"The environment variable {ex} must be set with a non-empty string") from ex

        return values


class CredentialsFile(HopprBaseSchemaModel):
    """
    Credentials file data model
    """

    kind: Literal["Credentials"]
    credential_required_services: list[CredentialRequiredService] = Field(
        [],
        unique_items=True,
        description=(
            "List of CredentialRequiredService objects to provide "
            "authentication to remote repositories and/or registries"
        ),
    )


CredentialsMap = Annotated[dict[str, CredentialRequiredService], Field(...)]


class Credentials(CredentialsFile):
    """
    CredentialsModel populated with environment variable passwords
    """

    lookup: ClassVar[CredentialsMap] = {}

    def __getitem__(self, item: str) -> CredentialRequiredService:
        return self.lookup[item]

    @root_validator(pre=True, allow_reuse=True)
    @classmethod
    def validate_credentials(cls, values: DictStrAny) -> DictStrAny:
        """
        Dynamically parse credentials using specified environment variables

        Args:
            values (DictStrAny): Field values of root object being validated

        Raises:
            AssertionError: Environment variable not set

        Returns:
            DictStrAny: The modified field values
        """
        for service in values.get("credential_required_services", []):
            cls.lookup[service.get("url")] = CredentialRequiredService.parse_obj(service)

        return values

    @classmethod
    def find(cls, url: str) -> CredentialRequiredService | None:
        """
        Return credentials for a repo URL
        """
        longest_matching_cred_url = ""

        for cred_url in cls.lookup:
            if cred_url in url and len(cred_url) > len(longest_matching_cred_url):
                longest_matching_cred_url = cred_url

        return cls.lookup.get(longest_matching_cred_url)

    @classmethod
    def load(cls, source: str | FilePath | DictStrAny | None) -> Credentials | None:
        """
        Load credentials file from local path or dict
        """
        # Can't use structural pattern matching here, pylint falsely reports unsubscriptable-object
        if source is None:
            return None

        if isinstance(source, dict):
            return cls.parse_obj(source)

        if isinstance(source, (str, Path)):
            return cls.parse_file(source)

        raise TypeError("'source' argument must be one of: 'str', 'Path', 'dict[str, Any]'")



"""
Tool for manipulating bundles for airgapped transfers.
"""
from __future__ import annotations

import hoppr_cyclonedx_models.cyclonedx_1_5 as cdx

from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process, hoppr_rerunner
from hoppr.exceptions import (
    HopprCredentialsError,
    HopprError,
    HopprExperimentalWarning,
    HopprLoadDataError,
    HopprPluginError,
)
from hoppr.logger import HopprLogger
from hoppr.models import HopprContext
from hoppr.models.affect import Affect
from hoppr.models.credentials import CredentialRequiredService, Credentials
from hoppr.models.licenses import (
    License,
    LicenseChoice,
    LicenseExpression,
    LicenseExpressionItem,
    LicenseMultipleItem,
    MultipleLicenses,
    NamedLicense,
    SPDXLicense,
)
from hoppr.models.manifest import Manifest
from hoppr.models.sbom import (
    Component,
    ComponentType,
    DataOutputType,
    EventType,
    ExternalReference,
    ExternalReferenceType,
    Hash,
    IssueType,
    LearningType,
    Metadata,
    PatchType,
    Property,
    Sbom,
    SubjectMatterType,
    Tools,
    Vulnerability,
)
from hoppr.models.transfer import ComponentCoverage, Transfer
from hoppr.models.types import BomAccess, PurlType
from hoppr.result import Result


__version__ = "1.11.5"

__all__ = [
    "__version__",
    "BomAccess",
    "ComponentCoverage",
    "CredentialRequiredService",
    "Credentials",
    "hoppr_process",
    "hoppr_rerunner",
    "HopprContext",
    "HopprCredentialsError",
    "HopprError",
    "HopprExperimentalWarning",
    "HopprLoadDataError",
    "HopprLogger",
    "HopprPlugin",
    "HopprPluginError",
    "Manifest",
    "PurlType",
    "Result",
    "Transfer",
    # CycloneDX model classes and type aliases
    "cdx",
    "Affect",
    "Component",
    "ComponentType",
    "DataOutputType",
    "EventType",
    "ExternalReference",
    "ExternalReferenceType",
    "Hash",
    "IssueType",
    "LearningType",
    "License",
    "LicenseChoice",
    "LicenseExpression",
    "LicenseExpressionItem",
    "LicenseMultipleItem",
    "Metadata",
    "MultipleLicenses",
    "NamedLicense",
    "PatchType",
    "Property",
    "Sbom",
    "SPDXLicense",
    "SubjectMatterType",
    "Tools",
    "Vulnerability",
]



"""Framework for manipulating bundles for airgapped transfers."""
from __future__ import annotations

import logging
import sys

from typing import TYPE_CHECKING

from typer import prompt

from hoppr.in_toto import generate_in_toto_layout
from hoppr.models.credentials import Credentials
from hoppr.models.manifest import Manifest
from hoppr.models.transfer import Transfer
from hoppr.processor import HopprProcessor


if TYPE_CHECKING:
    from pathlib import Path


def bundle(  # pylint: disable=too-many-arguments, too-many-locals, unused-argument
    manifest_file: Path,
    credentials_file: Path | None,
    transfer_file: Path,
    log_file: Path | None,
    verbose: bool = False,
    strict_repos: bool = True,
    create_attestations: bool = False,
    functionary_key_path: Path | None = None,
    functionary_key_prompt: bool = False,
    functionary_key_password: str | None = None,
    previous_delivery: Path | None = None,
    delivered_sbom: Path | None = None,
    basic_term: bool = False,
    experimental: bool = False,
    ignore_errors: bool = False,
):
    """Run the stages specified in the transfer config file on the content specified in the manifest."""
    log_level = logging.DEBUG if verbose else logging.INFO

    processor = HopprProcessor(
        transfer_file=transfer_file,
        manifest_file=manifest_file,
        credentials_file=credentials_file,
        create_attestations=create_attestations,
        functionary_key_path=functionary_key_path,
        functionary_key_password=functionary_key_password,
        log_level=log_level,
        log_file=log_file,
        strict_repos=strict_repos,
        previous_delivery=previous_delivery,
        ignore_errors=ignore_errors,
    )

    result = processor.run()

    if result.is_fail() and not ignore_errors:
        sys.exit(1)

    if delivered_sbom:
        delivered_sbom.write_text(processor.context.delivered_sbom.json(indent=2))


def generate_layout(
    transfer_file: Path,
    project_owner_key_path: Path,
    functionary_key_path: Path,
    project_owner_key_prompt: bool,
    project_owner_key_password: str,
):
    """Create in-toto layout based on transfer file."""
    if project_owner_key_prompt:
        project_owner_key_password = prompt(f"Enter password for {project_owner_key_path!s}", hide_input=True)

    transfer = Transfer.load(transfer_file)
    generate_in_toto_layout(transfer, project_owner_key_path, functionary_key_path, project_owner_key_password)


def validate(input_files: list[Path], credentials_file: Path, transfer_file: Path):
    """Validate multiple manifest files for schema errors."""
    if credentials_file is not None:
        Credentials.load(credentials_file)

    if transfer_file is not None:
        Transfer.load(transfer_file)

    for file in input_files:
        Manifest.load(file)



"""
Collector plugin for apt packages
"""
from __future__ import annotations

import fnmatch
import os
import shutil

from configparser import ConfigParser
from os import PathLike
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING, Mapping
from urllib.parse import urlparse

import jc

from pydantic import SecretStr

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import BatchCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_process, hoppr_rerunner
from hoppr.exceptions import HopprPluginError
from hoppr.models import HopprContext
from hoppr.models.credentials import Credentials
from hoppr.models.manifest import Repository
from hoppr.models.sbom import Component
from hoppr.models.types import PurlType
from hoppr.result import Result

if TYPE_CHECKING:
    from packageurl import PackageURL


class CollectAptPlugin(BatchCollectorPlugin):
    """
    Collector plugin for apt packages
    """

    required_commands: list[str] = ["apt", "apt-cache"]
    supported_purl_types: list[str] = ["deb"]
    products: list[str] = ["deb/*"]

    def __init__(
        self,
        context: HopprContext,
        config: dict | None = None,
        config_dir: str | PathLike[str] = Path.cwd() / ".hoppr-apt",
    ) -> None:
        super().__init__(
            context=context,
            config=config,
        )

        self.manifest_repos: list[str] = []

        self.manifest_repos = [f"{repo.url}" for repo in self.context.repositories[PurlType.DEB]]

        proxy_args = self._repo_proxy()
        self.config_dir = Path(config_dir)

        self.apt_paths_dict: Mapping[str, Mapping] = {
            "etc": {
                "apt": {"auth.conf": None, "preferences.d": {}, "sources.list": None, "sources.list.d": {}},
                "cache": {"apt": {"archives": {"lock": None, "partial": {}}}},
            },
            "var": {
                "lib": {
                    "apt": {"lists": {"lock": None, "partial": {}}},
                    "dpkg": {"lock-frontend": None, "status": None},
                }
            },
        }

        self.base_command = [
            self.required_commands[0],
            f"--option=Dir={self.config_dir}",
            f"--option=Dir::State::status={self.config_dir / 'var' / 'lib' / 'dpkg' / 'status'}",
            *proxy_args,
        ]

        self.debug_args = [
            "--option=Debug::pkgProblemResolver=1",
            "--option=Debug::Acquire::http=1",
            "--option=Debug::Acquire::https=1",
            "--option=Debug::Acquire::gpgv=1",
            "--option=Debug::pkgAcquire=1",
            "--option=Debug::pkgAcquire::Auth=1",
            "--option=Debug::pkgAcquire::Worker=1",
            "--option=Debug::pkgPackageManager=1",
            "--option=Debug::sourceList=1",
        ]

    def _artifact_string(self, purl: PackageURL) -> str:
        arch = purl.qualifiers.get("arch")
        return f"{purl.name}{f':{arch}' if arch else ''}{f'={purl.version}' if purl.version else ''}"

    def _download_component(self, purl: PackageURL) -> None:
        artifact = self._artifact_string(purl)

        try:
            # Download the component
            download_result = self.run_command(
                [
                    *self.base_command,
                    *(self.debug_args if self.get_logger().is_verbose() else []),
                    "download",
                    artifact,
                ]
            )

            download_result.check_returncode()
        except CalledProcessError as ex:
            raise ex

    def _get_component_download_info(self, purl: PackageURL) -> tuple[str, str]:
        artifact = self._artifact_string(purl=purl)

        try:
            # Try getting component download URL
            url_result = self.run_command(
                [
                    *self.base_command,
                    "--print-uris",
                    "download",
                    artifact,
                ]
            )

            url_result.check_returncode()

            if len(url_result.stdout.decode("utf-8")) == 0:
                raise HopprPluginError
        except (CalledProcessError, HopprPluginError) as ex:
            msg = f"Failed to get download URL for component: '{purl}'"
            self.get_logger().debug(msg=msg, indent_level=2)
            raise ex

        # Take the first URL and download filename if multiple are returned
        # Apt download output format:
        #   '<download URL>' <URL-encoded download filename> <download size> SHA512:<package hash>
        # Extract the output into a tuple of the form (<download URL>, <URL-encoded download filename>). For example:
        #   ("http://archive.ubuntu.com/ubuntu/pool/main/g/git/git_2.34.1-1ubuntu1.5_amd64.deb",
        #    "git_1%3a2.34.1-1ubuntu1.5_amd64.deb")
        result = url_result.stdout.decode("utf-8").split("\n")[0]
        found_url, download_filename = result.split(" ")[:2]

        self.get_logger().debug(msg=f"Found URL: {found_url}", indent_level=3)
        self.get_logger().debug(msg=f"Download filename: {download_filename}", indent_level=3)

        return found_url.strip("'"), download_filename

    def _get_download_url_path(self, purl: PackageURL) -> str:
        """
        Get the path segment of the component download URL using `apt-cache show`
        """
        artifact = self._artifact_string(purl)

        command = [*self.base_command, *(self.debug_args if self.get_logger().is_verbose() else []), "show", artifact]
        command[0] = "apt-cache"

        cmd_result = self.run_command(command)

        try:
            cmd_result.check_returncode()
        except CalledProcessError as ex:
            raise ex

        pkg_info = jc.parse(parser_mod_name="ini", data=cmd_result.stdout.decode(encoding="utf-8"))

        if not isinstance(pkg_info, dict) or not isinstance(pkg_info["Filename"], str):
            raise TypeError("Parsed output not in the expected format.")

        return pkg_info["Filename"]

    def _get_found_repo(self, found_url: str) -> str | None:
        """
        Identify the repository associated with the specified URL
        """
        return next((repo for repo in self.manifest_repos if found_url.startswith(repo)), None)

    def _populate_apt_folder_structure(self, apt_path: Path, path_dict: Mapping[str, Mapping | None]) -> None:
        for key, value in path_dict.items():
            # None type indicates file to create
            if value is None:
                apt_file = apt_path / key
                apt_file.touch(exist_ok=True, mode=0o644)
            # dict type indicates directory to create
            elif isinstance(value, dict):
                apt_dir = apt_path / key
                apt_dir.mkdir(exist_ok=True, parents=True)

                if len(value.items()) > 0:
                    self._populate_apt_folder_structure(apt_path=apt_dir, path_dict=value)
            else:
                raise TypeError("Value is not expected type.")

    def _populate_auth_conf(self, repo_list: list[Repository], file: Path) -> None:
        """
        Populate Apt authentication config file
        """
        creds: list[str] = []
        for repo in repo_list:
            repo_credentials = Credentials.find(f"{repo.url}")
            if repo_credentials is not None and isinstance(repo_credentials.password, SecretStr):
                creds.append(
                    f"machine {repo.url} "
                    f"login {repo_credentials.username} "
                    f"password {repo_credentials.password.get_secret_value()}\n"
                )

        # Set restrictive permissions on Apt authentication config file
        file.chmod(mode=0o600)

        with file.open(mode="w+", encoding="utf-8") as auth_conf:
            auth_conf.write("\n".join(creds))

    def _populate_sources_list(self, repo_list: list[str], file: Path) -> None:
        """
        Populate sources list
        """
        # Read data from /etc/os-release file
        parser = ConfigParser()

        with (Path("/") / "etc" / "os-release").open(mode="r", encoding="utf-8") as os_release:
            parser.read_string(string=f"[os-release]\n{os_release.read()}")

        version_codename = parser["os-release"]["VERSION_CODENAME"]

        sources: list[str] = []
        for repo in repo_list:
            for component in ["main restricted", "universe", "multiverse"]:
                sources.extend(
                    (
                        f"deb {repo} {version_codename} {component}",
                        f"deb {repo} {version_codename}-updates {component}",
                        f"deb {repo} {version_codename}-security {component}",
                    )
                )

        with file.open(mode="w+", encoding="utf-8") as sources_list:
            sources_list.write("\n".join(sources))

    def _repo_proxy(self) -> set[str]:
        proxy_args: list[str] = []

        for proto in ["http", "https"]:
            if proxy := os.getenv(f"{proto}_proxy"):
                proxy_args = [*proxy_args, f"--option=Acquire::{proto}::Proxy={proxy}"]

        no_proxy_urls = [item for item in os.getenv("no_proxy", "").split(",") if item != ""]

        for url in self.manifest_repos:
            parsed_url = urlparse(url)

            for pattern in no_proxy_urls:
                # Check if pattern is a substring or wildcard match of manifest repo URL
                if pattern in parsed_url.netloc or fnmatch.fnmatch(name=parsed_url.netloc, pat=pattern):
                    proxy_args = [
                        *proxy_args,
                        f"--option=Acquire::http::Proxy::{parsed_url.netloc}=DIRECT",
                        f"--option=Acquire::https::Proxy::{parsed_url.netloc}=DIRECT",
                    ]

        return set(proxy_args)

    def get_version(self) -> str:
        return __version__

    @hoppr_process
    def pre_stage_process(self) -> Result:
        self._populate_apt_folder_structure(apt_path=self.config_dir, path_dict=self.apt_paths_dict)

        self._populate_sources_list(
            repo_list=self.manifest_repos,
            file=self.config_dir / "etc" / "apt" / "sources.list",
        )

        self._populate_auth_conf(
            repo_list=self.context.repositories[PurlType.DEB],
            file=self.config_dir / "etc" / "apt" / "auth.conf",
        )

        if not self.context.strict_repos:
            system_apt_path = Path("/") / "etc" / "apt"
            plugin_apt_path = self.config_dir / "etc" / "apt"

            # Copy system Apt source lists into temporary directory
            shutil.copyfile(
                src=system_apt_path / "sources.list",
                dst=plugin_apt_path / "sources.list.d" / "system.list",
            )

            shutil.copytree(
                src=system_apt_path / "sources.list.d",
                dst=plugin_apt_path / "sources.list.d",
                dirs_exist_ok=True,
            )

        # Populate user Apt cache
        result = self.run_command(
            [
                *self.base_command,
                *(self.debug_args if self.get_logger().is_verbose() else []),
                "--option=Dir::Etc::trusted=/etc/apt/trusted.gpg",
                "--option=Dir::Etc::trustedparts=/etc/apt/trusted.gpg.d",
                "update",
            ]
        )

        if result.returncode != 0:
            return Result.fail("Failed to populate Apt cache.")

        return Result.success()

    @hoppr_rerunner
    def collect(self, comp: Component) -> Result:
        """
        Copy a component to the local collection directory structure
        """

        purl = hoppr.utils.get_package_url(comp.purl)

        try:
            found_url, download_filename = self._get_component_download_info(purl=purl)
        except (CalledProcessError, HopprPluginError) as ex:
            return Result.retry(message=str(ex))

        download_url_path = Path(self._get_download_url_path(purl=purl))
        subdir = download_url_path.parent

        if self.context.strict_repos:
            # Strip surrounding single quotes to compare against manifest repos
            repo = self._get_found_repo(found_url)

            # Return failure if found APT URL is not from a repo defined in the manifest
            if repo is None:
                return Result.fail(
                    f"Apt download URL does not match any repository in manifest. (Found URL: '{found_url}')"
                )

            result = self.check_purl_specified_url(purl, repo)
            if not result.is_success():
                return result
        else:
            # Default to found_url with the path component removed
            repo = found_url.removesuffix(str(download_url_path))

        target_dir = self.directory_for(purl.type, repo, subdir=str(subdir))

        try:
            self._download_component(purl=purl)
        except CalledProcessError:
            msg = f"Failed to download Apt artifact {purl.name} version {purl.version}"
            return Result.retry(msg)

        self.get_logger().info(msg="Moving downloaded component:", indent_level=2)
        self.get_logger().info(msg=f"source: {download_filename}", indent_level=3)
        self.get_logger().info(msg=f"destination: {target_dir}", indent_level=3)

        shutil.move(src=download_filename, dst=target_dir)

        self.set_collection_params(comp, repo, target_dir)

        return Result.success(return_obj=comp)



"""
Collector plugin for npm packages
"""
from __future__ import annotations

from urllib.parse import quote_plus
from pydantic import SecretStr

import requests

from requests.auth import HTTPBasicAuth

import hoppr.utils

from hoppr import __version__
from hoppr.base_plugins.collector import SerialCollectorPlugin
from hoppr.base_plugins.hoppr import hoppr_rerunner
from hoppr.models import HopprContext
from hoppr.models.credentials import CredentialRequiredService
from hoppr.models.sbom import Component
from hoppr.models.types import RepositoryUrl
from hoppr.result import Result


class CollectNpmPlugin(SerialCollectorPlugin):
    """
    Collector plugin for npm packages
    """

    supported_purl_types: list[str] = ["npm"]
    products: list[str] = ["npm/*"]
    system_repositories: list[str] = ["https://registry.npmjs.org"]

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context=context, config=config)

    def get_version(self):
        return __version__

    @hoppr_rerunner
    def collect(self, comp: Component, repo_url: str, creds: CredentialRequiredService | None = None) -> Result:
        """
        Download npm package via get request
        """
        purl = hoppr.utils.get_package_url(comp.purl)
        self.get_logger().info(msg=f"Copying npm package from {comp.purl}", indent_level=2)

        tar_name = f"{purl.name}{f'-{purl.version}' if purl.version else ''}.tgz"
        npm_url = RepositoryUrl(url=repo_url) / quote_plus(purl.namespace or "") / purl.name / "-" / tar_name

        authentication = None
        if creds is not None and isinstance(creds.password, SecretStr):
            authentication = HTTPBasicAuth(creds.username, creds.password.get_secret_value())

        response = requests.get(npm_url.url, auth=authentication, timeout=300, proxies=self._get_proxies_for(repo_url))

        if response.status_code not in range(200, 300):
            msg = f"Failed to locate npm package for {comp.purl}, return_code={response.status_code}"
            self.get_logger().debug(msg=msg, indent_level=2)

            return Result.retry(message=msg)

        target_dir = self.directory_for(purl.type, repo_url, subdir=purl.namespace)
        package_out_path = target_dir / tar_name
        package_out_path.write_bytes(response.content)

        self.set_collection_params(comp, repo_url, target_dir)
        return Result.success(return_obj=comp)
